import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from datetime import date
import os
import concurrent.futures

import time
from cProfile import Profile
from pstats import Stats

import matplotlib
#
# matplotlib.use('TkAgg')

matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['font.size'] = 13
matplotlib.rcParams['font.serif'] = 'Helvetica'

matplotlib.rcParams['font.weight'] = 'bold'

matplotlib.rcParams['axes.labelweight'] = 'bold'
matplotlib.rcParams['axes.labelsize'] = 'large'

matplotlib.rcParams['legend.fontsize'] = 'large'

matplotlib.rcParams['axes.titleweight'] = 'bold'
matplotlib.rcParams['text.latex.preamble'] = [r'\usepackage{sfmath} \boldmath']

matplotlib.rcParams['xtick.labelsize'] = 'large'
matplotlib.rcParams['ytick.labelsize'] = 'large'

matplotlib.rcParams['lines.markersize'] = 10.0

plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 3))
plt.ticklabel_format(style='sci', axis='x')

print(matplotlib.rcParams.keys())


# for debug
np.random.seed(0)


class Bi_graph_exp:
    def __init__(self, n_s=20, n_t=50, t_MC=1000, k_A=5, k_B=5, T=500, theta=10000, oracle='greedy',
                 S_B_type='max', G_type=None, edge_prob_type='indegree', gamma=2.5):
        # source node num
        self.n_s = n_s
        # target node num
        self.n_t = n_t
        # number of repetition to calc inf
        self.t_MC = t_MC
        # number of SA
        self.k_A = k_A
        # number of SB
        self.k_B = k_B
        self.edge_prob_type = edge_prob_type
        if G_type is None:
            self.G_type = "complete"
            # adj matrix
            self.G = np.zeros((n_s, n_t))
            # first 10 nodes has higher edge probability
            self.G[0:10] = np.random.rand(10, n_t)
            self.G[10:] = np.random.rand(self.n_s-10, n_t) * 0.1
            # graph structure mask edge_mask(s,t) = 1 if (s,t) is an edge
            self.edge_mask = np.zeros((n_s, n_t)) + 1
        elif G_type == 'powerlaw':
            self.G_type = G_type
            self.G = np.zeros((n_s, n_t))
            self.edge_mask = np.zeros((n_s, n_t))
            self.build_powerlaw_bi_graph(gamma)
        elif G_type == 'real':
            self.G_type = G_type
            self.build_real_world_bi_graph()
        self.T = T
        # randomly choose k_B competitors
        self.S_B_type = S_B_type
        if S_B_type == 'uniform':
            self.S_B = np.random.choice(self.n_s, self.k_B, replace=False)
        elif S_B_type == 'max':
            self.S_B = np.array([])
            self.S_B, _ = self.greedy(self.G)
            self.S_B = np.array(self.S_B)
        # number of RR sets
        self.theta = theta
        # RR set or MC greedy
        self.oracle = oracle
        self.dir = f'result/{date.today().isoformat()}'
        if not os.path.isdir(self.dir):
            os.mkdir(self.dir)
        self.log_file = f'{self.dir}/{self.oracle}_{self.G_type}_{self.n_s}_{self.n_t}_{self.T}'

    def build_powerlaw_bi_graph(self, gamma=2.5):
        k = 5
        deg_seq = np.array(nx.random_powerlaw_tree_sequence(self.n_s, gamma=gamma, seed=0, tries=1000000))
        # smaller index has larger degree
        sq = k * np.sort(deg_seq)[::-1]
        print(np.sum(sq))
        print(sq)
        # edge connection mask
        for i_s, d in enumerate(sq):
            target_nodes = np.random.choice(self.n_t, d, replace=False)
            # p = 1/indegree
            for target_node in target_nodes:
                # self.G[i_s, target_node] = (k-1)/d
                self.edge_mask[i_s, target_node] = 1
        if self.edge_prob_type == 'indegree':
            # in_degree
            in_degs = np.sum(self.edge_mask, axis=0)
            print(in_degs)
            one_over_in_degs = np.array([0 if deg==0 else 1.0/deg for deg in in_degs])
            one_over_in_degs = one_over_in_degs.reshape(1, self.n_t)
            one_over_in_degs = np.repeat(one_over_in_degs, self.n_s, axis=0)
            self.G = self.edge_mask * one_over_in_degs
        else:
            self.G[0:10] = np.random.rand(10, self.n_t)
            self.G[10:] = np.random.rand(self.n_s - 10, self.n_t) * 0.1
            self.G = self.G * self.edge_mask


    def preprocess_real_data(self):
        s_nodes = {}
        t_nodes = {}
        s_id = 0
        t_id = 0
        edge_list = {}
        with open('data/ydata-ysm-advertiser-bids-v1_0.txt', 'r') as f:
            for line in f:
                s = line.split('\t')[1]
                t = line.split('\t')[2]
                if s not in s_nodes:
                    s_nodes[s] = s_id
                    edge_list[s_id] = set()
                    s_id += 1
                if t not in t_nodes:
                    t_nodes[t] = t_id
                    t_id += 1
                edge_list[s_nodes[s]].add(t_nodes[t])

        print(f"source: {len(s_nodes)}, target: {len(t_nodes)}, edges: {sum([len(edge_list[k]) for k in edge_list])}")
        with open('data/yahoo_ad.txt', 'w') as f:
            f.write(f'{len(s_nodes)} {len(t_nodes)}\n')
            for s in edge_list:
                for t in edge_list[s]:
                    f.write(f'{s} {t}\n')

    def build_real_world_bi_graph(self):
        line_id = 0
        with open('data/yahoo_ad.txt', 'r') as f:
            for line in f:
                if line_id == 0:
                    self.n_s = int(line.split()[0])
                    self.n_t = int(line.split()[1])
                    self.edge_mask = np.zeros((self.n_s, self.n_t))
                    line_id += 1
                else:
                    s = int(line.split()[0])
                    t = int(line.split()[1])
                    self.edge_mask[s, t] = 1
                    line_id +=1
        if self.edge_prob_type == 'indegree':
            # in_degree
            in_degs = np.sum(self.edge_mask, axis=0)
            print(in_degs)
            one_over_in_degs = np.array([0 if deg==0 else 1.0/deg for deg in in_degs])
            one_over_in_degs = one_over_in_degs.reshape(1, self.n_t)
            one_over_in_degs = np.repeat(one_over_in_degs, self.n_s, axis=0)
            self.G = self.edge_mask * one_over_in_degs
        else:
            self.G = np.zeros((self.n_s, self.n_t))
            self.G[0:10] = np.random.rand(10, self.n_t)
            self.G[10:] = np.random.rand(self.n_s - 10, self.n_t) * 0.1
            self.G = self.G * self.edge_mask

    def cal_inf_mc(self, G, S_A):
        sum_A = 0
        # Monte Carlo simulation
        for _ in range(self.t_MC):
            inf_A = np.zeros(self.n_t)
            inf_B = np.zeros(self.n_t)
            # A's inference spread
            for i in S_A:
                inf_A += (np.random.rand(self.n_t) < G[i]).astype(int)
            # B's inference spread
            for i in self.S_B:
                inf_B += (np.random.rand(self.n_t) < G[i]).astype(int)
            # B dominates A
            sum_A += np.count_nonzero(inf_A + inf_B) - np.count_nonzero(inf_B)
        return sum_A / self.t_MC

    def cal_inf(self, G, S_A):
        if len(self.S_B) == 0:
            inf_A = np.sum((1 - np.prod(1 - G[S_A], axis=0)))
        else:
            inf_A = np.sum(np.prod(1 - G[self.S_B], axis=0) * (1 - np.prod(1 - G[S_A], axis=0)))
        return inf_A

    # greedily select seed set of A
    def greedy(self, G):
        S_A = []
        S = self.S_B.tolist()
        for _ in range(self.k_A):
            inf_out = np.zeros(self.n_s)
            for i in range(self.n_s):
                if i not in S:
                    inf_out[i] = self.cal_inf(G, S_A + [i])
            #                 print('node '+str(i)+', inf '+str(inf_out[i]))
            i_max = np.argmax(inf_out)
            S_A.append(i_max)
            S.append(i_max)
        #         print('select node '+str(i_max))
        #         print('inf '+str(inf_out.max()))
        return S_A, inf_out.max()

    # Reverse Reachable set
    def RR(self, G):
        start_time = time.time()
        S_A = np.zeros(self.k_A, dtype=np.int16) - 1
        R = np.zeros((self.theta, self.n_s))
        for i in range(self.theta):
            r = np.random.randint(self.n_t)
            g = (np.random.rand(self.n_s) < G[:, r])
            if np.any(g[self.S_B]):
                continue
            else:
                R[i] = g.astype(int)
        print(f"RIS: {time.time() - start_time}")
        n_covered = 0
        for i in range(self.k_A):
            if np.count_nonzero(R) == 0:
                tmp = np.random.randint(self.n_s)
                while tmp in S_A or tmp in self.S_B:
                    tmp = np.random.randint(self.n_s)
                S_A[i] = tmp
                continue
            i_max = np.argmax(np.sum(R, axis=0))
            R_del = (R[:, i_max] == 1)
            n_covered += np.count_nonzero(R_del)
            R[R_del] = 0
            S_A[i] = i_max
        spread = n_covered / self.theta * self.n_t
        print("--- %s seconds ---" % (time.time() - start_time))
        return S_A, spread

    def TS(self, log_ind=0, write_per_step=1):
        start_time = time.time()

        np.random.seed(log_ind)
        cum_reg = 0.0
        inf_t = np.zeros(self.T)
        f_write = open(f'{self.log_file}_ep_0.0_er_ts_{self.edge_prob_type}_{self.S_B_type}_{log_ind}', 'w')

        Q_beta = np.ones((self.n_s, self.n_t, 2))


        # approximate solution of the offline model

        if self.oracle == 'greedy':
            S_A_offline, opt = self.greedy(self.G)
            print(f"oracle: {S_A_offline}, opt: {opt}")
        else: # oracle == 'RR'
            S_A_offline, _ = self.RR(self.G)
            opt = self.cal_inf(self.G, S_A_offline)
            print(f"oracle: {S_A_offline}, opt: {opt}")


        for t in range(self.T):
            start_round = time.time()

            G_t = np.random.beta(Q_beta[:,:,0], Q_beta[:,:,1])
            # assume we know the graph structure
            G_t = G_t * self.edge_mask

            if self.oracle == 'greedy':
                S_A, _ = self.greedy(G_t)
            else:  # oracle == 'RR'
                S_A, _ = self.RR(G_t)

            # fix bug in rho
            S_all = np.append(S_A, self.S_B)
            for i in range(self.n_s):
                if i in S_all:
                    X_i = (np.random.rand(self.n_t) < self.G[i]).astype(int)
                    Q_beta[i,:,0] = Q_beta[i,:,0] + X_i
                    Q_beta[i,:,1] = Q_beta[i,:,1] + 1 - X_i

            # expected reward for this round
            print(f"chosen S_A: {S_A}")
            inf_t[t] = self.cal_inf(self.G, S_A)
            # print(f"mc: {self.cal_inf_mc(self.G, S_A)}, truth: {self.cal_inf(self.G, S_A)}")
            cum_reg += opt-inf_t[t]
            print('round ' + str(t) + ', reward: ' + str(inf_t[t]))
            if t%write_per_step == 0:
                f_write.write(f'{t} {inf_t[t]} {opt-inf_t[t]} {cum_reg}\n')
            end_round = time.time()
            print(f"time: {end_round-start_round}")
        reg_t = opt - inf_t
        f_write.close()
        end_time = time.time()
        print(f"total time: {end_time-start_time}, avg_time: {(end_time-start_time)/self.T}")
        return reg_t, opt

    def cucb(self, explore_ratio=1., ep_greedy=0., etc=0, log_ind=0, write_per_step=10):
        start_time = time.time()


        np.random.seed(log_ind)
        cum_reg = 0.0
        if etc == 0:
            f_write = open(f'{self.log_file}_ep_{ep_greedy}_er_{explore_ratio}_{self.edge_prob_type}_{self.S_B_type}_{log_ind}', 'w')
        else:
            f_write = open(
                f'{self.log_file}_ep_{ep_greedy}_er_{explore_ratio}_etc_{etc}_{self.edge_prob_type}_{self.S_B_type}_{log_ind}',
                'w')
        mu_hat = np.ones((self.n_s, self.n_t))
        rho = np.ones((self.n_s, self.n_t))
        T_i = np.zeros((self.n_s, self.n_t))
        inf_t = np.zeros(self.T)

        # approximate solution of the offline model

        if self.oracle == 'greedy':
            S_A_offline, opt = self.greedy(self.G)
            print(f"oracle: {S_A_offline}, opt: {opt}")
        else: # oracle == 'RR'
            S_A_offline, _ = self.RR(self.G)
            opt = self.cal_inf(self.G, S_A_offline)
            print(f"oracle: {S_A_offline}, opt: {opt}")


        for t in range(self.T):
            start_round = time.time()
            UCB = np.clip(mu_hat + explore_ratio * rho, a_min=0, a_max=1)
            LCB = np.clip(mu_hat - explore_ratio * rho, a_min=0, a_max=1)
            G_t = UCB
            G_t[self.S_B] = LCB[self.S_B]
            # assume we know the graph structure
            G_t = G_t * self.edge_mask
            # ep_greedy

            if etc==0:
                if np.random.rand() > ep_greedy:
                    if self.oracle == 'greedy':
                        S_A, _ = self.greedy(G_t)
                    else:  # oracle == 'RR'
                        S_A, _ = self.RR(G_t)
                else:
                    # S_A \ S_B
                    S_A_candidate = list(set(range(self.n_s)) - set(self.S_B))
                    S_A = np.random.choice(S_A_candidate, self.k_A, replace=False)
            else:
                if t < etc:
                    S_A_candidate = list(set(range(self.n_s)) - set(self.S_B))
                    S_A = np.random.choice(S_A_candidate, self.k_A, replace=False)
                elif t == etc:
                    if self.oracle == 'greedy':
                        S_A, _ = self.greedy(G_t)
                    else:  # oracle == 'RR'
                        S_A, _ = self.RR(G_t)
                    S_A_continue = S_A
                else:
                    S_A = S_A_continue
                    inf_t[t] = reward_continue
                    cum_reg += opt - inf_t[t]
                    print('round ' + str(t) + ', reward: ' + str(inf_t[t]))
                    f_write.write(f'{t} {inf_t[t]} {opt - inf_t[t]} {cum_reg}\n')
                    continue
            # fix bug in rho
            S_all = np.append(S_A, self.S_B)
            for i in range(self.n_s):
                if i in S_all:
                    T_i[i] += 1
                    X_i = (np.random.rand(self.n_t) < self.G[i]).astype(int)
                    mu_hat[i] = mu_hat[i] + (X_i - mu_hat[i]) / T_i[i]
                    rho[i] = np.sqrt(3 * np.log(t + 1) / (2 * T_i[i]))
                else:
                    if T_i[i,0] != 0:
                        rho[i] = np.sqrt(3 * np.log(t + 1) / (2 * T_i[i]))

            # expected reward for this round
            print(f"chosen S_A: {S_A}")
            inf_t[t] = self.cal_inf(self.G, S_A)
            if t == etc:
                reward_continue = inf_t[t]
            # print(f"mc: {self.cal_inf_mc(self.G, S_A)}, truth: {self.cal_inf(self.G, S_A)}")
            cum_reg += opt-inf_t[t]
            print('round ' + str(t) + ', reward: ' + str(inf_t[t]))
            if t%write_per_step == 0:
                f_write.write(f'{t} {inf_t[t]} {opt-inf_t[t]} {cum_reg}\n')
            end_round = time.time()
            print(f"time: {end_round-start_round}")
        reg_t = opt - inf_t
        f_write.close()
        end_time = time.time()
        print(f"total time: {end_time-start_time}, avg_time: {(end_time-start_time)/self.T}")
        return reg_t, opt


    def profile_prog(self):
        # for optimization
        profiler = Profile()
        profiler.runcall(self.cucb)
        stats = Stats(profiler)
        stats.strip_dirs()
        stats.sort_stats('cumulative')
        stats.print_stats()

    def plot(self, explore_ratio=1., ep_greedy=0., etc=0, plot_range_low=0,
             plot_range_high=1, present_round=2000, color='b', marker='d-'):
        line_id = 0
        step = 200
        n_trials = plot_range_high - plot_range_low
        round = [i for i in range(present_round)]
        regret = np.zeros((n_trials, present_round))
        S_type = 'uniform'
        if etc == 0:
            f_name_prefix = f'result/2020-06-01/greedy_real_1000_10475_500_ep_{ep_greedy}_er_{explore_ratio}_indegree_{S_type}'
        else:
            f_name_prefix = f'result/2020-06-11/greedy_real_1000_10475_40000_ep_{ep_greedy}_er_{explore_ratio}_etc_{etc}_indegree_{S_type}'

        for i in range(plot_range_low, plot_range_high):
            # with open(f'{self.log_file}_ep_{ep_greedy}_er_{explore_ratio}_{self.edge_prob_type}_{i}', 'r') as f:
            # with open(f'result/2020-05-27/2020-05-27/RR_powerlaw_200_500_1000_ep_{ep_greedy}_er_{explore_ratio}_{i}', 'r') as f:
            # with open(f'result/2020-05-27/2020-05-27/RR_powerlaw_100_1000_3000_ep_{ep_greedy}_er_{explore_ratio}_indegree_{i}', 'r') as f:
            # with open(f'result/2020-05-28/greedy_powerlaw_300_1000_3000_ep_{ep_greedy}_er_{explore_ratio}_indegree_{i}', 'r') as f:
            # with open(f'result/2020-05-29/greedy_real_1000_10475_3000_ep_{ep_greedy}_er_{explore_ratio}_indegree_max_{i}', 'r') as f:
            with open(f'{f_name_prefix}_{i}', 'r') as f:
            # with open(
            #         f'result/2020-06-01/greedy_powerlaw_300_1000_2000_ep_{ep_greedy}_er_{explore_ratio}_indegree_uniform_{i}',
            #         'r') as f:
                for line in f:
                    if line_id == present_round -1:
                        line_id = 0
                        break
                    r = int(line.split()[0])
                    if r % 10 != 0:
                        continue
                    reg = float(line.split()[3])
                    if i == 0:
                        round[line_id] = r
                    regret[i, line_id] = reg
                    line_id += 1
        round = np.array(round)
        regret_mean = np.array([np.mean(regret[:, round_id]) for round_id in range(present_round)])
        regret_std = np.array([np.std(regret[:, round_id]) for round_id in range(present_round)])

        regret_mean = np.append(regret_mean[:step:int(step/4)], regret_mean[step::step-1])
        regret_std = np.append(regret_std[:step:int(step/4)], regret_std[step::step-1])
        round = np.append(round[:step:int(step/4)], round[step::step-1])

        # names
        if etc == 0:

            if explore_ratio == 0 and ep_greedy == 0:
                name = 'EMP'
            if ep_greedy > 0:
                name = r"$\epsilon$-Greedy ($\epsilon={}$)".format(ep_greedy)
            if explore_ratio > 0:
                name = r"OCIM-OIFU ($\alpha_{{\rho}} = {}$)".format(explore_ratio)
        else:
            name = r"OCIM-ETC ($ETC = {}$)".format(etc)


        plt.plot(round, regret_mean, f'{color}{marker}', label=name)
        plt.fill_between(round, regret_mean - 1.96 * regret_std/np.sqrt(n_trials),
                         regret_mean + 1.96 * regret_std/np.sqrt(n_trials), color=color, alpha=0.1)




if __name__ == "__main__":
    n_trials = 1
    t = 500
    cut = 10
    n_etc = [int(t / cut * (i + 1)) for i in range(8)]
    print(n_etc)
    exp = Bi_graph_exp(n_s=1000, n_t=10574, theta=10000, t_MC=300, T=t,
                       k_B=5, k_A=5, oracle='greedy', G_type='real', edge_prob_type='indegree', S_B_type='max')
    # exp = Bi_graph_exp(n_s=300, n_t=10, theta=10000, t_MC=1000, T=500,
    #                    k_B=5, k_A=5, oracle='greedy', G_type='powerlaw', edge_prob_type='indegree', S_B_type='max')
    with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
        for i in range(n_trials):
    #         for j in n_etc:
    #             executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, etc=j, log_ind=i, write_per_step=10)
    #
            # TS
    #        executor.submit(exp.TS, log_ind=i)
    #         # cucb
    #         executor.submit(exp.cucb, explore_ratio=1.0, ep_greedy=0.0, log_ind=i)
    #         # exp.cucb(explore_ratio=1.0, ep_greedy=0.0, log_ind=i)
    #
    #         # ep_greedy 0.05
    #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.05, log_ind=i)
    #         # exp.cucb(explore_ratio=0.0, ep_greedy=0.05)
    #
    #         # ep_greedy 0.01
    #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.01, log_ind=i)
    #         # exp.cucb(explore_ratio=0.0, ep_greedy=0.01)
    #
    #         # empirical_mean
    #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, log_ind=i)
    #         # exp.cucb(explore_ratio=0.0, ep_greedy=0.0)
    #
    #         # gamma 0.1
    #         executor.submit(exp.cucb, explore_ratio=0.1, ep_greedy=0.0, log_ind=i)
    #
    #         # gamma 0.2
    #         executor.submit(exp.cucb, explore_ratio=0.2, ep_greedy=0.0, log_ind=i)
    #
    #         # gamma 0.05
             executor.submit(exp.cucb, explore_ratio=0.05, ep_greedy=0.0, log_ind=i)
    #
    #         # gamma 0.01
    #         executor.submit(exp.cucb, explore_ratio=0.01, ep_greedy=0.0, log_ind=i)
    # #         # etc 50
    # #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, etc=50, log_ind=i)
    # #
    # #         # etc 100
    # #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, etc=100, log_ind=i)
    # #
    # #         # etc 150
    # #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, etc=150, log_ind=i)
    # #
    # #         # etc 200
    # #         executor.submit(exp.cucb, explore_ratio=0.0, ep_greedy=0.0, etc=200, log_ind=i)



    #
    # #exp = Bi_graph_exp(n_s=100, n_t=1000, theta=10000, t_MC=1000, T=5,
    # #                    k_B=4, k_A=5, oracle='RR', G_type='powerlaw', log_ind=0)
    # # reg_t, opt = exp.cucb(ep_greedy=0.00, explore_ratio=0.1)
    # colors = ["b", "r", "g", "y", "m", "c"]
    # markers = ["*--", "P--", "s--", "d:", "o-", "^-"]
    # # for i in range(5):
    # #     etc = 10 * i + 100
    # #     exp.plot(ep_greedy=0.0, explore_ratio=0.0, plot_range_low=0, etc=etc, plot_range_high=n_trials, color=colors[i],
    # #              marker=markers[i])
    # # exp.plot(ep_greedy=0.05, explore_ratio=0.0, plot_range_low=0, plot_range_high=n_trials, color=colors[5],
    # #          marker=markers[5])
    # exp.plot(ep_greedy=0.0, explore_ratio=0.0, plot_range_low=0, etc=10000, plot_range_high=n_trials, color=colors[0],
    #                       marker=markers[0])
    # # exp.plot(ep_greedy=0.05, explore_ratio=0.0, plot_range_low=0, plot_range_high=n_trials, color=colors[3], marker=markers[3])
    # # exp.plot(ep_greedy=0.01, explore_ratio=0.0, plot_range_low=0, plot_range_high=n_trials, color=colors[4], marker=markers[4])
    # # exp.plot(ep_greedy=0.0, explore_ratio=0.0, plot_range_low=0, plot_range_high=n_trials, color=colors[5], marker=markers[5])
    # # exp.plot(ep_greedy=0.0, explore_ratio=0.05, plot_range_low=0, plot_range_high=n_trials, color=colors[0], marker=markers[0])
    # # exp.plot(ep_greedy=0.0, explore_ratio=0.1, plot_range_low=0, plot_range_high=n_trials, color=colors[1], marker=markers[1])
    # # exp.plot(ep_greedy=0.0, explore_ratio=0.2, plot_range_low=0, plot_range_high=n_trials, color=colors[2], marker=markers[2])
    # plt.xlabel("Round T")
    # plt.ylabel("Regret")
    # plt.legend()
    # # plt.show()
    # # # plt.savefig('figure/100_1000_powerlaw_RR')
    # # plt.savefig('figure/200_500_powerlaw_random_RR')
    # # plt.savefig('figure/300_1000_powerlaw_random_greedy')
    # plt.savefig('figure/2020-06-11_10000_10475_powerlaw_indegree_uniform_greedy_ETC')
    # # # print(exp.S_B)
    # # # print(opt)
    # # # plt.plot(np.cumsum(reg_t))
    # # # plt.show()


